// Kod rdowy programu dla cyklu artykuw pt. Wasny silnik graficzny.
// Autor: Robert Kraus
//
// monte carlo path tracer (a)

#include <iostream>
#include <cstdlib>
#include <limits>
#include <cmath>
#include <vector>
#include <string>
#include <GL/glut.h>
#include <IL/il.h>
#include <IL/ilu.h>

///// stae globalne
////////////////////////////////////////////////////////////////////////////////

std::numeric_limits<float> real;
const float epsilon3 = 0.001f;   // staa dla bdw zaokrgle numerycznych
const float epsilon5 = 0.00001f; // staa dla bdw zaokrgle numerycznych
const float infinity = real.infinity();  // nieskoczono
const float undefined = real.infinity(); // warto niezdefiniowana
const float Pi     = 3.14159265f;
const float PiMul2 = 6.28318530f;
typedef unsigned int nat;            // typ liczb naturalnych
#define nt(x) ((nat)(x))             // rzutowanie na liczby naturalne
#define fl(x) ((float)(x))           // rzutowanie na liczby rzeczywiste
#define bt(x) ((unsigned char)(x))   // rzutowanie na liczby {0..255}
#define isNull(ptr) ( NULL == ptr )  // czy wskanik jest pusty
#define notNull(ptr) ( NULL != ptr ) // czy wskanik jest niepusty
// liczba losowa z przedziau [0,1] (rozkad jednostajny)
inline float uRand() { return fl(rand()) / fl(RAND_MAX); } 

///// wektor 2-wymiarowy
////////////////////////////////////////////////////////////////////////////////

struct vec2 {
  typedef const vec2& i; // input ref
  float u, v;

  vec2(float u, float v) : u(u), v(v) {}
  vec2() { u = v = 0.0f; }
};

// dugo wektora 2d
inline float length(vec2::i w)
{  return sqrt(w.u * w.u + w.v * w.v);  }

///// operacje arytmetyczne dla wektora 2d

inline vec2 operator+ (vec2::i a, vec2::i b)
{  return vec2(a.u + b.u, a.v + b.v);  }

inline vec2 operator- (vec2::i a, vec2::i b)
{  return vec2(a.u - b.u, a.v - b.v);  }

inline vec2 operator* (vec2::i a, float t)
{  return vec2(a.u * t, a.v * t);  }

///// wektor 3-wymiarowy
////////////////////////////////////////////////////////////////////////////////

struct vec3 {
  typedef const vec3& i; // input ref
  typedef vec3& io;

  float x, y, z;

  vec3() { x = y = z = 0.0f; }
  vec3(float x2, float y2, float z2) : x(x2), y(y2), z(z2) {}
  vec3(float x2, float y2, float z2, float len) : x(x2), y(y2), z(z2) {
    float newLen = len / sqrt(x*x + y*y + z*z);
    x *= newLen;  y *= newLen;  z *= newLen;
  }
};

///// operacje arytmetyczne dla wektora 3d

inline vec3 operator+ (vec3::i a, vec3::i b)
{ return vec3(a.x+b.x, a.y+b.y, a.z+b.z); }

inline void operator+= (vec3::io a, vec3::i b)
{ a = a + b; }

inline vec3 operator- (vec3::i a, vec3::i b)
{ return vec3(a.x-b.x, a.y-b.y, a.z-b.z); }

inline void operator-= (vec3::io a, vec3::i b)
{ a = a - b; }

inline vec3 operator* (vec3::i a, float t)
{ return vec3(a.x*t, a.y*t, a.z*t); }

inline vec3 operator/ (vec3::i a, float t)
{ return vec3(a.x/t, a.y/t, a.z/t); }

///// koniec operacji arytmetycznych

// przeksztacenie wektora u na wektor -u
inline vec3 invert(vec3::i u)
{ return vec3(-u.x,-u.y,-u.z); }

// przeksztacenie wektora u na wektor o dugoci 1
inline vec3 unitise(vec3::i u)
{ return vec3(u.x, u.y, u.z, 1.0f); }

// dugo wektora 3d
inline float length(vec3::i v)
{ return sqrt(v.x*v.x + v.y*v.y + v.z*v.z); }

// odlego pomidzy punktami p1 i p2
inline float distance(vec3::i p1, vec3::i p2)
{ return length(p2 - p1); }

// iloczyn skalarny wektorw v1 i v2
inline float dotProd(vec3::i v1, vec3::i v2)
{ return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z; }

// iloczyn wektorowy wektorw v1 i v2
// wynikiem jest wektor prostopady do wektorw v1 i v2
inline vec3 uCrossProd(vec3::i v1, vec3::i v2) {
  return vec3(
    v1.y*v2.z - v2.y*v1.z,
    v1.z*v2.x - v2.z*v1.x,
    v1.x*v2.y - v2.x*v1.y,
    1.0f
  );
}

// przeniesienie wektora V do przestrzeni zbudowanej na wektorach Ex, Ey, Ez
// oraz jego normalizacja
inline vec3 uChangeSpace(vec3::i V, vec3::i Ex, vec3::i Ey, vec3::i Ez) {
  return vec3(
    V.x*Ex.x + V.y*Ey.x + V.z*Ez.x,
    V.x*Ex.y + V.y*Ey.y + V.z*Ez.y,
    V.x*Ex.z + V.y*Ey.z + V.z*Ez.z,
    1.0f
  );
}

///// promie
////////////////////////////////////////////////////////////////////////////////

struct ray {
  typedef const ray& i; // input ref
  vec3 origin;
  vec3 abackDir;
  vec3 aheadDir;

  ray(vec3 origin, vec3 aheadDir2) :
    origin(origin), aheadDir(aheadDir2), abackDir(invert(aheadDir2)) {}
};

///// kolor
////////////////////////////////////////////////////////////////////////////////

// przycinanie wartoci ujemnych do zera (typ float)
#define CUT(x) ( x < 0.0f ? 0.0f : x )

struct color {
  typedef const color& i; // input ref
  typedef color& io; // input/output ref
  float r, g, b;

  color() { r = g = b = 0.0f; }
  color(float r, float g, float b) : r(r), g(g), b(b) {}
  color(float v) { r = g = b = CUT(v); }
};

color                           // kolory:
  black(0.0f),                  //   czarny
  grey(0.9f),                   //   szary
  white(1.0f),                  //   biay
  red(0.95f, 0.32f, 0.23f),     //   czerwony
  green(0.71f, 0.89f, 0.1f),    //   zielony
  blue(0.19f, 0.77f, 0.98f),    //   niebieski
  darkBlue(0.04f, 0.42f, 0.5f), //   ciemny niebieski
  yellow(1.0f, 0.98f, 0.74f);   //   ty

// sprawdzanie czy kolor ma przynajmniej jeden niezerowy kana
inline bool isBlack(color::i c)
{ return (c.r + c.g + c.b > 0.0f) ? false : true;  }

///// operacje arytmetyczne dla kolorw

inline color operator+ (color::i a, color::i b)
{ return color(a.r + b.r, a.g + b.g, a.b + b.b);  }

inline void operator+= (color::io a, color::i b)
{ a = a + b;  }

inline color operator- (color::i a, color::i b)
{ return color(a.r - b.r, a.g - b.g, a.b - b.b);  }

inline color operator* (color::i a, color::i b)
{ return color(a.r * b.r, a.g * b.g, a.b * b.b);  }

inline color operator* (color::i a, float t)
{ return color(a.r * t, a.g * t, a.b * t);  }

inline color operator/ (color::i a, float c)
{ return color(a.r / c, a.g / c, a.b / c);  }

///// teksturowanie
////////////////////////////////////////////////////////////////////////////////

// mapa kolorw
struct colorMap {
  color* map; // tablica kolorw
  nat xRes, yRes, size; // wymiary tekstury
  // map[x + y*xRes] = teksel o wsprzdnych (x,y)

  colorMap() : map(NULL), xRes(0), yRes(0), size(0) {}
  colorMap(color* map, nat xRes, nat yRes) :
    map(map), xRes(xRes), yRes(yRes), size(xRes*yRes) {}

  // prbkowanie punktowe
  //color operator() (float u, float v) const {
  //  nat ui = nt(u * fl(xRes));
  //  nat vi = nt(v * fl(yRes));
  //  return map[ (ui + vi*xRes) % size ];
  //}

  // prbkowanie z filtrowaniem dwuliniowym
  color operator() (float u, float v) const {
    // x = u * fl(xRes)
    // y = v * fl(yRes)
    nat ui = nt(u * fl(xRes)); // x0
    nat vi = nt(v * fl(yRes)); // y0
    nat a = (  ui    +  vi    * xRes ) % size; // (x0,y0)
    nat b = ( (ui+1) +  vi    * xRes ) % size; // (x1,y0)
    nat c = (  ui    + (vi+1) * xRes ) % size; // (x0,y1)
    nat d = ( (ui+1) + (vi+1) * xRes ) % size; // (x1,y1)
    // g'(x) = f(x0,y0) + ( f(x1,y0) - f(x0,y0) ) * ((x-x0) / (x1-x0))
    color ab = map[a] + (map[b]-map[a]) * ( u*fl(xRes) - fl(ui) );
    // g''(x) = f(x0,y1) + ( f(x1,y1) - f(x0,y1) ) * ((x-x0) / (x1-x0))
    color cd = map[c] + (map[d]-map[c]) * ( u*fl(xRes) - fl(ui) );
    // g(x,y) = g'(x) + (g''(x) - g'(x)) * ((y-y0) / (y1-y0))
    return ab + (cd - ab) * ( v*fl(yRes) - fl(vi) );
  }
};

// mapa wektorw
struct vectorMap {
  vec3* map; // tablica wektorow
  nat xRes, yRes, size; // wymiary tekstury
  // map[x + y*xRes] = teksel o wsprzdnych (x,y)

  vectorMap() : map(NULL), xRes(0), yRes(0), size(0) {}
  vectorMap(vec3* map, nat xRes, nat yRes)
    : map(map), xRes(xRes), yRes(yRes), size(xRes*yRes) {}

  // prbkowanie punktowe
  //vec3 operator() (float u, float v) const {
  //  nat ui = nt(u * fl(xRes));
  //  nat vi = nt(v * fl(yRes));
  //  return map[ (ui + vi*xRes) % size ];
  //}

  // prbkowanie z filtrowaniem dwuliniowym
  vec3 operator() (float u, float v) const {
    // x = u * fl(xRes)
    // y = v * fl(yRes)
    nat ui = nt(u * fl(xRes)); // x0
    nat vi = nt(v * fl(yRes)); // y0
    nat a = (  ui    +  vi    * xRes ) % size; // (x0,y0)
    nat b = ( (ui+1) +  vi    * xRes ) % size; // (x1,y0)
    nat c = (  ui    + (vi+1) * xRes ) % size; // (x0,y1)
    nat d = ( (ui+1) + (vi+1) * xRes ) % size; // (x1,y1)
    // g'(x) = f(x0,y0) + ( f(x1,y0) - f(x0,y0) ) * ((x-x0) / (x1-x0))
    vec3 ab = map[a] + (map[b]-map[a]) * ( u*fl(xRes) - fl(ui) );
    // g''(x) = f(x0,y1) + ( f(x1,y1) - f(x0,y1) ) * ((x-x0) / (x1-x0))
    vec3 cd = map[c] + (map[d]-map[c]) * ( u*fl(xRes) - fl(ui) );
    // g(x,y) = g'(x) + (g''(x) - g'(x)) * ((y-y0) / (y1-y0))
    return ab + (cd - ab) * ( v*fl(yRes) - fl(vi) );
  }
};

ILuint ImgId;
ILinfo ImgInfo;
ILenum Error;
ILubyte* Image;

// wczytywanie mapy kolorw
colorMap* loadColorMap(std::string name) {
  // inicjalizacja
  ilInit();
  ilGenImages(1, &ImgId);
  ilBindImage(ImgId);
  // wczytanie obrazu
  ilLoadImage((wchar_t*)name.c_str()); // pod g++ naley usun (wchar_t*)
  iluGetImageInfo(&ImgInfo);
  Image = ilGetData();

  // przepisanie obrazu z tablicy bajtw na tablic trjek float'w
  nat xRes = ilGetInteger(IL_IMAGE_WIDTH);
  nat yRes = ilGetInteger(IL_IMAGE_HEIGHT);
  printf("preparing texture: %s @ %d x %d", name.c_str(), xRes, yRes);
  color* map = new color[xRes*yRes];
  for (nat y = 0; y < yRes; y++)
    for (nat x = 0; x < xRes; x++) {
      nat i = x+y*xRes;
      map[i] = color(
        fl(Image[3*i+0])/255.0f,
        fl(Image[3*i+1])/255.0f,
        fl(Image[3*i+2])/255.0f
      );
    }

  ilDeleteImages(1, &ImgId);
  printf(" [done]\n");
  return new colorMap(map, xRes, yRes);
}

// wczytywanie mapy wektorw
vectorMap* loadVectorMap(std::string name) {
  // inicjalizacja
  ilInit();
  ilGenImages(1, &ImgId);
  ilBindImage(ImgId);
  // wczytanie obrazu
  ilLoadImage((wchar_t*)name.c_str()); // pod g++ naley usun (wchar_t*)
  iluGetImageInfo(&ImgInfo);
  Image = ilGetData();

  // przepisanie obrazu z tablicy bajtw na tablic trjek float'w
  nat xRes = ilGetInteger(IL_IMAGE_WIDTH);
  nat yRes = ilGetInteger(IL_IMAGE_HEIGHT);
  printf("preparing texture: %s @ %d x %d", name.c_str(), xRes, yRes);
  vec3* map = new vec3[xRes*yRes];
  for (nat y = 0; y < yRes; y++)
    for (nat x = 0; x < xRes; x++) {
      nat i = x + y*xRes;
      nat k = x + (yRes-1-y)*xRes;
      map[k] = vec3(
        (fl(Image[3*i+0])/255.0f -0.5f) * 2.0f,  
         fl(Image[3*i+2])/255.0f,
        (fl(Image[3*i+1])/255.0f -0.5f) * 2.0f,
         1.0f
      );
    }

  ilDeleteImages(1, &ImgId);
  printf(" [done]\n");
  return new vectorMap(map, xRes, yRes);
}

///// materia
////////////////////////////////////////////////////////////////////////////////

struct material {
  color dr;        // diffuse reflectance (odbicie rozproszone)
  colorMap* drMap; // diffuse reflectance map
  color e;         // emittance (emisja wiata)
  vectorMap* normalMap; // mapa wektorw normalnych

  bool isDiffuse;       // czy materia odbija z rozproszeniem
  bool hasTextures;     // czy materia posiada tekstury
  bool hasDrMap;        // czy materia posiada map odbicia rozproszonego
  bool hasNormalMap;    // czy materia posiada map wektorw normalnych

  material(color::i dr, color::i e) :
    dr(dr), e(e), drMap(NULL), normalMap(NULL) {
      isDiffuse = !isBlack(dr);
      hasTextures = hasDrMap = hasNormalMap = false;
    }

  material(material m, colorMap* cMap, vectorMap* nMap) :
    dr(m.dr), drMap(cMap), normalMap(nMap) {
      isDiffuse = m.isDiffuse;
      hasDrMap = notNull(drMap) ? true : false;
      hasNormalMap = notNull(normalMap) ? true : false;
      hasTextures = hasDrMap || hasNormalMap;
    }
};

material
  whiteFloor(white, black),        // biaa podoga
  sky(black, yellow),
  emitter(black, yellow * 12.0f),  // powierzchnia wiecca
  redStone(red, black),            // czerwony kamie
  blueStone(blue, black),          // niebieski kamie
  yellowStone(yellow, black);      // ty kamie

///// obiekty geometryczne
////////////////////////////////////////////////////////////////////////////////

// sfera
struct sphere {
  typedef const sphere& i; // input ref
  vec3 center;   // rodek sfery
  float radius;  // promie sfery
  material* mat; // wskanik na materia

  sphere(vec3 center, float radius, material* m) :
    center(center), radius(radius), mat(m) {}
};

// trjkt
struct triangle {
  typedef const triangle& i; // input ref

  vec3 v1, v2, v3;    // wierzchoki trjkta
  vec2 uv1, uv2, uv3; // wierzchoki trjkta w przestrzeni tekstury
  vec3 N;             // wektor normalny do paszczyzny trjkta
  float d;            // rwnanie paszczyzny: dotProd(N,(x,y,z)) + d = 0
  material* mat;      // wskanik na materia

  triangle(
    vec3::i v1,
    vec3::i v2,
    vec3::i v3,
    vec2::i uv1,
    vec2::i uv2,
    vec2::i uv3,
    material* m
  ) :
    v1(v1), v2(v2), v3(v3), uv1(uv1), uv2(uv2), uv3(uv3), mat(m) {
    N = uCrossProd(v2-v1,v3-v1);
    d = -dotProd(N, v1);
  }
};

///// wektory normalne
////////////////////////////////////////////////////////////////////////////////

inline vec3 normal(sphere::i s, vec3::i p)
{  return (p - s.center) / s.radius;  }

inline vec3 normal(triangle::i t)
{  return t.N;  }

///// teksturowanie
////////////////////////////////////////////////////////////////////////////////

// pole trjkta rozpitego w przestrzeni 3d na wierzchokach v1, v2, v3
// implementacja wzoru Herona
inline float area(vec3::i v1, vec3::i v2, vec3::i v3) {
  float a = length(v2 - v1);
  float b = length(v3 - v1);
  float c = length(v3 - v2);
  float p = (a + b + c) / 2.0f;
  return sqrt(p*(p-a)*(p-b)*(p-c));
}

// pole trjkta rozpitego w przestrzeni 2d na wierzchokach v1, v2, v3
// implementacja wzoru Herona
inline float area(vec2::i v1, vec2::i v2, vec2::i v3) {
  float a = length(v2 - v1);
  float b = length(v3 - v1);
  float c = length(v3 - v2);
  float p = (a + b + c) / 2.0f;
  return sqrt(p*(p-a)*(p-b)*(p-c));
}

// przeksztacenie punktu z przestrzeni sfery na przestrze tekstury
inline vec2 mapSurfaceToTexture(sphere::i s, vec3::i p) {
  vec3 lp = p - s.center;

  float yDIVr = lp.y / s.radius;
  float v = 
    ( abs(yDIVr - 1.0f) < epsilon5 ) ? 0.0f : // v = 0 (grny biegun)
    ( abs(yDIVr + 1.0f) < epsilon5 ) ? 1.0f : // v = 1 (dolny biegun)
    acos(yDIVr) / Pi;                         // bez osobliwoci

  float rMULsin = s.radius * sin(Pi*v);
  float u = // 1,2 -> biegun // 3,4 -> |x/rMULsin|=1 -> acos nieokrelony
    ( abs(v) < epsilon5 ) ? 0.0f :               // 1) (u,v) = (0.0, 0)
    ( abs(v - 1.0f) < epsilon5 ) ? 0.0f :        // 2) (u,v) = (0.0, 1)
    ( abs(lp.x - rMULsin) < epsilon5 ) ? 0.0f :  // 3) (u,v) = (0.0, v) 
    ( abs(lp.x + rMULsin) < epsilon5 ) ? 0.5f :  // 4) (u,v) = (0.5, v)
    acos(lp.x / rMULsin) / PiMul2;               // bez osobliwoci

  return vec2(u, v);
}

// przeksztacenie punktu z przestrzeni trjkta na przestrze tekstury
inline vec2 mapSurfaceToTexture(triangle::i t, vec3::i p) {
  float abc = area(t.v1 , t.v2 , t.v3);
  float   r = area(  p  , t.v2 , t.v3) / abc;
  float   s = area(t.v1 ,   p  , t.v3) / abc;
  float   q = 1.0f - r - s;
  return t.uv1*r + t.uv2*s + t.uv3*q;
}

// przeksztacenie punktu z przestrzeni tekstury na przestrze sfery
inline vec3 mapTextureToSurface(sphere::i s, vec2::i p) {
  return vec3(
    s.center.x + s.radius * sin(Pi * p.v) * cos(PiMul2 * p.u),
    s.center.y + s.radius * cos(Pi * p.v),
    s.center.z + s.radius * sin(Pi * p.v) * sin(PiMul2 * p.u)
  );
}

// przeksztacenie punktu z przestrzeni tekstury na przestrze trjkta
inline vec3 mapTextureToSurface(triangle::i t, vec2::i p) {
  float abc = area(t.uv1 , t.uv2 , t.uv3);
  float   r = area(  p   , t.uv2 , t.uv3) / abc;
  float   s = area(t.uv1 ,   p   , t.uv3) / abc;
  float   q = 1.0f - r - s;
  return t.v1*r + t.v2*s + t.v3*q;
}

///// przecicia
////////////////////////////////////////////////////////////////////////////////

struct hit {
  float t;  // odlego trafienia od rda promienia
  vec3 ip;  // punkt przecicia
  vec3 N;   // wektor normalny do powierzchni trafionego obiektu
  vec3 Ex;  // wektor na osi X w przestrzeni rozpitej na wektorach Ex, N, Ez
  vec3 Ez;  // wektor na osi Z w przestrzeni rozpitej na wektorach Ex, N, Ez
  vec2 uv;  // wsprzdne trafienia w przestrzeni tekstury
  material* mat; // wskanik na materia
  
  hit() : t(undefined) {}

  hit(sphere::i S, float t, ray::i r) : t(t) {
    if (t == undefined)  return;
    ip = r.origin + r.aheadDir * t;  // punkt przecicia
    mat = S.mat;                     // wskanik na materia
    N = normal(S, ip);               // normalna do powierzchni
    if ( ! mat -> hasTextures ) return;
    uv = mapSurfaceToTexture(S, ip); // punkt ip w przestrzeni tekstury
    if ( ! mat -> hasNormalMap ) return;
    Ex = mapTextureToSurface(S, vec2(uv.u + epsilon3, uv.v)) - ip;
    Ex = unitise(Ex);
    Ez = mapTextureToSurface(S, vec2(uv.u, uv.v + epsilon3)) - ip;
    Ez = unitise(Ez);
  }

  hit(triangle::i T, float t, ray::i r) : t(t) {
    if (t == undefined) return;
    ip = r.origin + r.aheadDir * t;  // punkt przecicia
    mat = T.mat;                     // wskanik na materia
    N = normal(T);                   // normalna do powierzchni
    if ( ! mat -> hasTextures ) return;
    uv = mapSurfaceToTexture(T, ip); // punkt ip w przestrzeni tekstury
    if ( ! mat -> hasNormalMap ) return;
    Ex = mapTextureToSurface(T, vec2(uv.u + epsilon3, uv.v)) - ip;
    Ex = unitise(Ex);
    Ez = mapTextureToSurface(T, vec2(uv.u, uv.v + epsilon3)) - ip;
    Ez = unitise(Ez);
  }
};

// przecicie ze sfer
inline float intersect(sphere::i S, ray::i r) {
  vec3 v = r.origin - S.center;
  float b = dotProd(v, r.aheadDir);
  float d2 = b*b - dotProd(v,v) + S.radius*S.radius;
  if (d2 < 0.0f) return undefined;
  float d = sqrt(d2);
  float t1 = -b-d, t2 = -b+d;
  if (t1 > epsilon5) return t1;
  if (t2 > epsilon5) return t2;
  return undefined;
}

// przecicie z trjktem
inline float intersect(triangle::i T, ray::i r) {
  float t = (-T.d - dotProd(T.N, r.origin)) / dotProd(T.N, r.aheadDir);
  if ( !(t > epsilon5 && t < infinity) )
    return undefined;
  vec3 ip = r.origin + r.aheadDir * t;
  float abc = area(T.v1 , T.v2 , T.v3);
  float pbc = area( ip  , T.v2 , T.v3);
  float apc = area(T.v1 ,  ip  , T.v3);
  float abp = area(T.v1 , T.v2 ,  ip );
  if ( pbc + apc + abp > abc + epsilon3 )
    return undefined;
  return t;
}

///// kamera
////////////////////////////////////////////////////////////////////////////////

struct camera {
  vec3 location;          // pooenie kamery
  vec3 up, front, right;  // wektory kierunkw: do gry, do przodu, w prawo
  nat xRes, yRes;         // rozdzielczo obrazu

  camera(vec3 location, vec3 up, vec3 front, nat xRes, nat yRes) :
    location(location),
    up(up),
    front(front),
    right(uCrossProd(front,up)),
    xRes(xRes),
    yRes(yRes)
    { }
};

// odchylenie poziome od rodka obrazu/obiektywu
inline float xA(camera& c, float x)
{  return (2.0f * x / fl(c.xRes) -1.0f);  }

// odchylenie pionowe od rodka obrazu/obiektywu
inline float yA(camera& c, float y)
{  return (2.0f * y / fl(c.yRes) -1.0f);  }

///// scena
////////////////////////////////////////////////////////////////////////////////

struct scene {
  std::vector<sphere> spheres; // zbir sfer 
  std::vector<triangle> triangles; // zbir trjktw
  camera cam; // kamera

  scene(camera c) : cam(c) {
    // wczytanie tekstur
    colorMap* tex1 = loadColorMap("ice.png");

    // tworzenie materiaw zalenych od tekstur
    material* rStone = new material(redStone, tex1, NULL);
    material* bStone = new material(blueStone, tex1, NULL);

    // tworzenie obiektw sceny
    triangles.push_back(triangle(
      vec3(-4.0f, 0.0f,  2.0f),  // v1
      vec3( 4.0f, 0.0f,  2.0f),  // v2
      vec3( 0.0f, 0.0f, -4.0f),  // v3
      vec2(0.0f, 0.0f),          // v1 na teksturze
      vec2(0.5f, 1.0f),          // v2 na teksturze
      vec2(1.0f, 0.0f),          // v3 na teksturze
      &whiteFloor
    ));
    spheres.push_back(sphere(vec3(-0.8f,0.5f, 0.0f), 0.4f, rStone));
    spheres.push_back(sphere(vec3(-0.4f,1.0f, 0.5f), 0.2f, &emitter));
    spheres.push_back(sphere(vec3( 0.0f,0.2f, 0.4f), 0.1f, &emitter));
    spheres.push_back(sphere(vec3( 0.5f,1.3f,-0.2f), 0.2f, &emitter));
    spheres.push_back(sphere(vec3( 0.8f,0.5f, 0.0f), 0.4f, bStone));
  }
};

///// szukanie najbliszego przecicia
////////////////////////////////////////////////////////////////////////////////

// szukanie najbliszej sfery
inline hit nearestSphere(ray::i r, scene& s) {
  float nearest(undefined);  // najblisze znalezione przecicie
  nat id(0);                 // indeks najbliszej sfery
  for (nat i = 0; i < s.spheres.size(); i++) { // przegldamy wszystkie sfery
     float t = intersect(s.spheres[i], r); // obliczenie odlegoci
     if (t < nearest) { // czy obiekt jest bliej od ostatniego najbliszego
       nearest = t;
       id = i;
     }
  }
  return (nearest != infinity) ? hit(s.spheres[id], nearest, r) : hit();
}

// szukanie najbliszego trjkta
inline hit nearestTriangle(ray::i r, scene& s) {
  float nearest(undefined); // najblisze znalezione przecicie
  nat id(0);                // indeks najbliszego trjkta
  for (nat i = 0; i < s.triangles.size(); i++) { // przegldamy wszystkie trjkty
     float t = intersect(s.triangles[i], r); // obliczenie odlegoci
     if (t < nearest) { // czy obiekt jest bliej od ostatniego najbliszego
       nearest = t;
       id = i;
     }
  }
  return (nearest != infinity) ? hit(s.triangles[id], nearest, r) : hit();
}

// szukanie najbliszego obiektu
inline hit nearestHit(ray::i r, scene& s) {
  hit hitT = nearestTriangle(r, s);
  hit hitS = nearestSphere(r, s);
  return hitT.t < hitS.t ? hitT : hitS;
}

///// generowanie promieni
////////////////////////////////////////////////////////////////////////////////

// promie z kamery przez wsprzdne na paszczynie ekranu (x,y)
inline ray primaryRay(camera& c, float x, float y) { 
  return ray(
    c.location, // rdem jest pooenie kamery
    unitise(c.up * yA(c,y) + c.right * xA(c,x) + c.front)
  );
}

// losowy wekotr na psferze o promieniu 1
// rozkad cosinusa
inline vec3 sampleHemiSphere() {
  float Rd = uRand();
  float Ad = uRand();
  float c = sqrt(1.0 - Rd);
  return vec3(c * cos(PiMul2*Ad), sqrt(Rd), c * sin(PiMul2*Ad), 1.0f);
}

// odbicie z rozproszeniem
inline ray scatter(ray::i r, vec3::i p, vec3::i N) {
  vec3 Ex = uCrossProd(r.abackDir, N);
  vec3 Ez = uCrossProd(Ex, N);
  vec3 hsV = sampleHemiSphere();
  return ray(p, uChangeSpace(hsV, Ex, N, Ez));
}

///// rekurencyjne ledzenie promieni
////////////////////////////////////////////////////////////////////////////////

// r - ledzony promie
// s - scena, w ktrej ledzony jest promie
// d - gboko rekursji
color traceRay(ray::i r, scene& s, nat d) {
  if (d == 0) return black; // koniec rekursji
  hit h = nearestHit(r, s); // informacje o najbliszym przeciciu
  if (h.t == undefined) return black; // brak przecicia

  material& m = *h.mat; // materia trafionego obiektu
  color texMapped = notNull(m.drMap) ? (*m.drMap)(h.uv.u, h.uv.v) : white;
  if (m.hasNormalMap) // perturbacja wektora normalnego zgodnie z tekstur
    h.N = uChangeSpace((*m.normalMap)(h.uv.u, h.uv.v), h.Ex, h.N, h.Ez);

  if (m.isDiffuse) // odbicie rozproszone
    return m.dr * texMapped * traceRay(scatter(r, h.ip, h.N), s, d-1);
  else // emisja
    return m.e;
}

///// antyaliasing
////////////////////////////////////////////////////////////////////////////////

// s - scena, w ktrej ledzimy promienie
// (x,y) - wsprzdne piksela
// maxD - maksymalna gboko rekursji dla ledzenia promieni
// numS - ilo prbek (tj. ilo promieni pierwotnych przypadajcych na 1 piksel)
color splicer(scene& s, float x, float y, nat maxD, nat numS) {
  color f;
  for (nat i = 0; i < numS; i++)
    f += traceRay(
      primaryRay( s.cam, x + 0.5f -uRand(), y + 0.5f -uRand() ),
      s,
      maxD
    );

  return f / numS;
}

///// mapowanie tonw
////////////////////////////////////////////////////////////////////////////////

#define FILTER(x) ( x > 1.0f ? 1.0f : x )

void toneMapping(color::io c) {
  c.r = FILTER(c.r);
  c.g = FILTER(c.g);
  c.b = FILTER(c.b);
}

///// rendering
////////////////////////////////////////////////////////////////////////////////

// maksymalna gboko ledzenia promieni
nat maxTraceDepth = 4;

// ilo prbek (promieni pierwotnych) na 1 piksel
// im wiksza warto tym lepsza jako obrazu
nat numSamples = 1; 

void render(scene& s, GLubyte* picture) {
  camera& cam(s.cam);
  color pixel;
  for (nat y = 0; y < cam.yRes; y++)
    for (nat x = 0; x < cam.xRes; x++) {
      pixel = splicer(s, fl(x), fl(y), maxTraceDepth, numSamples);
      toneMapping(pixel);
      picture[3*(y*cam.xRes + x) +0] = bt(pixel.r * 255.0f);
      picture[3*(y*cam.xRes + x) +1] = bt(pixel.g * 255.0f);
      picture[3*(y*cam.xRes + x) +2] = bt(pixel.b * 255.0f);
    }
}

///// wywietlanie okna
////////////////////////////////////////////////////////////////////////////////

GLubyte* screen;
nat globalXRes, globalYRes;

void display() {
  glClear(GL_COLOR_BUFFER_BIT);
  glRasterPos2i(0, 0);
  glDrawPixels(globalXRes, globalYRes, GL_RGB, GL_UNSIGNED_BYTE, screen);
  glutSwapBuffers();
  glutPostRedisplay();
}

void reshape(int w, int h) {
  glViewport(0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0, (GLdouble) w, 0.0, (GLdouble) h);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

///// obsuga klawiatury
////////////////////////////////////////////////////////////////////////////////

scene* myScene;

inline void highQualityRender(nat n) {
  nat oldNumSamples = numSamples;
  numSamples = n;
  render(*myScene, screen);
  numSamples = oldNumSamples;
  glutPostRedisplay();
}

static void standard_keys(unsigned char key, int x, int y) {
  camera& cam = myScene -> cam;
  bool hqRendering = false;
  switch(key) {
    // poruszanie kamer
    case 'n': cam.location -= cam.up * 0.04; break;
    case 'h': cam.location += cam.up * 0.04; break;
    case 'm': cam.location += cam.right * 0.04; break;
    case 'b': cam.location -= cam.right * 0.04; break;
    case 'g': cam.location += cam.front * 0.04; break;
    case 'j': cam.location -= cam.front * 0.04; break;

    case '1':
      hqRendering = true;
      highQualityRender(100);
      break;
    case '2':
      hqRendering = true;
      highQualityRender(250);
      break;
    case '3':
      hqRendering = true;
      highQualityRender(500);
      break;
    case '4':
      hqRendering = true;
      highQualityRender(750);
      break;
    case '5':
      hqRendering = true;
      highQualityRender(1000);
      break;
    case '6':
      hqRendering = true;
      highQualityRender(5000);
      break;
    case '7':
      hqRendering = true;
      highQualityRender(10000);
      break;
    case '8':
      hqRendering = true;
      highQualityRender(50000);
      break;

    // zmiana iloci prbek (promieni pierwotnych)
    case 'q': printf("samples: %d\n", ++numSamples); break;
    case 'w': if (numSamples > 1) printf("samples: %d\n", --numSamples); break;

    // zmiana maksymalnej gbokoci ledzenia promieni
    case 'e': printf("depth: %d\n", ++maxTraceDepth); break;
    case 'r': if (maxTraceDepth > 1) printf("depth: %d\n", --maxTraceDepth); break;
  }

  if ( !hqRendering ) {
    render(*myScene, screen);
    glutPostRedisplay();
  }
}

///// main
////////////////////////////////////////////////////////////////////////////////

int main(int argc, char* argv[]) {
  // rozdzielczo wyjciowego obrazu
  globalXRes = 512;
  globalYRes = 512;
  printf("output resolution: %d x %d\n", globalXRes, globalYRes);

  // tworzenie sceny
  myScene = new scene(camera(
    vec3(0.0f, 0.3f, 1.5f),        // pooenie
    vec3(0.0f, 1.0f, 0.0f, 1.0f),  // kierunek do gry od pooenia
    vec3(0.0f, 0.0f, -1.0f, 1.0f), // kierunek do przodu od pooenia
    globalXRes,
    globalYRes
  ));

  // tworzenie tablicy pikseli oraz rendering
  screen = new GLubyte[myScene->cam.xRes * myScene->cam.yRes * 3];
  render(*myScene, screen);

  // inicjalizacja GLUT'a i utworzenie okna wywietlajcego obraz
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
  glutInitWindowSize(globalXRes, globalYRes);
  glutCreateWindow("simple raytracer");
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(standard_keys);

  glutMainLoop();
  return 0;
}
